/*
                                ENVSET.C

  Program to set specified environment variable(s) to a string(s) supplied
from standard input. This is a utility to be used from within a batch
file to capture a value and place it into the environment for subsequent
use during processing.

                   Compiled under Zortech C++ using:

                            ZTC -mci -o sev

       (full optimizations, no floating point, compact mem model)

                                SYNTAX:

                      SEV varname [varname . . .]

  Where varname is the name of the environment variable to set.  More
than one can be set per call.  The program obtains the values for these
variables as strings by reading standard input.  This means that you can
redirect output from a program into an environment variable, or you can
enter it from the keyboard.

  The most obvious use for SEV is as an aid during batch processing.
EXAMPLE:
                         @echo off
      +----------------> break | SEV brkstatus
      |                  if "%brkstatus%"=="BREAK is OFF" break on
      |                  ...
      |
    This will pipe the output from the BREAK inquiry into an environment
variable called 'brkstatus', where it can subsequently be tested.

  Another way to use SEV is to provide a menu of choices onscreen, and
capture the user's response in an environment variable.
EXAMPLE:
                         @echo off
                         :getans
                         cls
                         echo 1. PacMan
                         echo 2. SpaceWars
                         echo 3. Family Feud
                         echo    Select 1, 2, or 3
                         SEV answer
                         if "%ANSWER%"=="1" goto dopacman
                         if "%ANSWER%"=="2" goto dospacewars
                         if "%ANSWER%"=="3" goto dofeud
                         goto getans
                         ...

  SEV changes all copies of the environment, not just the newest one.
This is in contrast to the DOS set command. SEV locates the DOS master
environment by using an undocumented DOS interrupt, int 21h 52h.  This
interrupt finds the first MCB (Memory Control Block) allocated when
CONFIG.SYS is loaded.  This is used as a starting point from which to
skip to the first loading of the command interpreter, which will contain
the DOS master environment.  The master is then acquired by SEV, as well
as all subsequent environment blocks.

  Once the environment blocks have been found, standard input is scanned
for the values to be assigned to the variables specified as arguments to
SEV.  The environment records are modified to reflect the new variable
values, "sorted", and written back to memory.

  Thanks to:
                        Tom Gilbert's Heart&Mind
                         7127 Lafayette Avenue
                         Kansas City, KS 66109
                             (913) 299-2701

for the assembly code demonstrating the way to get at the MCBs.  For the
full set of assembly source files dealing with TSR management, including MARK,
UNMARK, FMARK, SHOWTSRS, MENUTSRS, FREERAM and WATCH, download

                               TSRW31.ARC

from CompuServe's IBMSYS forum, lib 3 -- 103,424 bytes.

                          John H. Staelens Jr.
                         CompuServe 73527,3273
                            Sun  05-13-1990
*/




#include            <stdio.h>
#include            <stdlib.h>
#include            <dos.h>
#include            <process.h>
#include            <string.h>
#include            <direct.h>
#define             SUCCEEDED           1
#define             FAILED              -1
#define             true                1
#define             false               0
#define             MAXLINE             128
#define             UCHAR               unsigned char




                                        /*************************************/
                                        /*  This class holds the strings     */
                                        /*  extracted from the environment   */
                                        /*  blocks as a linked list.         */
                                        /*************************************/
class slist {
    char *var;
    slist *next;
  public :                              /* Define constructor that sets up  */
                                        /*  defaults.                       */
    slist() {
      var = NULL;
      next = NULL;
    } /* end function: "slist()" */
                                        /* Now define data access functions */
    slist *getnext() {
      return next;
    } /* end function: "*getnext()" */
    void setnext(slist *nxt) {
      next = nxt;
    } /* end function: "setnext(slist *nxt)" */

    char *getvar() {
      return var;
    } /* end function: "*getvar()" */
    void setvar(char *fn) {
      var = strdup(fn);
    } /* end function: "setvar(char *fn)" */

                                        /* This next allows the sort routine*/
                                        /* to reset pointers rather than    */
                                        /* reallocating new strings through */
                                        /* the setvar method.               */
    void setvptr(char *fn) {
      var = fn;
    } /* end function: "setvptr(char *fn)" */
};




                                        /*************************************/
                                        /* This class of object holds all the*/
                                        /* data needed to rebuild an envir-  */
                                        /* onment block, including its loc-  */
                                        /* ation in memory, content, size,   */
                                        /* and usage.  These objects can be  */
                                        /* linked into a list, so that all   */
                                        /* copies of the environment can be  */
                                        /* acquired and updated.             */
                                        /*************************************/
class envrec : public slist {
    void *memloc;
    unsigned envsize;
    unsigned envused;
    slist *env_vars;
    int wcount;
    char invocation[MAXLINE];
    envrec *next;
  public :                              /* Define constructor that sets up  */
                                        /*  defaults.                       */
    envrec(void *ml = NULL, unsigned es = 128, unsigned eu = 0,
           slist *ev = NULL, int wc = 0, char *inv = NULL, envrec *nxt = NULL) {
      memloc = ml;
      envsize = es;
      envused = eu;
      env_vars = ev;
      wcount = wc;
      if (inv) strncpy(invocation, inv, MAXLINE);
      next = nxt;
    } /* end function: "envrec(void *ml = NULL, unsigned es = 128, unsigned eu = 0, slist *ev = NULL, int wc = 0, char *inv = NULL, envrec *nxt = NULL)" */
                                        /* Now define data access functions */
    void setloc(void *ml) {
      memloc = ml;
    } /* end function: "setloc(void *ml)" */
    void *getloc() {
      return memloc;
    } /* end function: "*getloc()" */

    void setsize(unsigned sz) {
      envsize = sz;
    } /* end function: "setsize(unsigned sz)" */
    unsigned getsize() {
      return envsize;
    } /* end function: "getsize()" */

                                        /* Next one used to assure that room*/
                                        /* is available to add new vars to  */
                                        /* an environment block.  It is the */
                                        /* only way provided for setting the*/
                                        /* 'envused' value, and should be   */
                                        /* called after all other data is   */
                                        /* has been initialized.            */
    void compute_used() {
      unsigned      count               = 0;
      slist         *tsl                = env_vars;
      
                                        /* First find string usage          */
      while (tsl) {
        count += strlen(tsl->getvar()) + 1;
        tsl = tsl->getnext();
      } /* end "while (tsl)" */
      count += 3;
                                        /* Adjust for zero byte + word count*/
                                        /* count following strings          */
      count += strlen(invocation) + 1;
                                        /* include the invocation line      */
      envused = count;
    } /* end function: "compute_used()" */
    unsigned getused() {
      return envused;
    } /* end function: "getused()" */

                                        /* Find out if the var is already   */
                                        /* present.  If so, just modify it. */
    void addvar(char *ev) {
      int           eq_pos_ev           = strcspn(ev, "=");
      int           done                = 0;
      slist         *tslist             = env_vars;
      
      if (tslist) {
        while (!done) {
          if (strncmp(tslist->getvar(), ev, eq_pos_ev) == 0) done = 1;
          else if (tslist->getnext()) tslist = tslist->getnext();
          else {
            tslist->setnext(new slist);
            tslist = tslist->getnext();
            done = 1;
          } /* end else clause of "if (tslist->getnext())" */
        } /* end "while (!done)" */
      } /* end "if (tslist)" */
      else {
        tslist = new slist;
        env_vars = tslist;
      } /* end else clause of "if (tslist)" */
                                        /* If this is a reset of the env var*/
                                        /* then null the string to be added */
      if (strlen(ev) == eq_pos_ev + 1) ev[0] = '\0';
      tslist->setvar(ev);
    } /* end function: "addvar(char *ev)" */

    slist *getlist() {
      return env_vars;
    } /* end function: "*getlist()" */

    void setwc(int wc) {
      wcount = wc;
    } /* end function: "setwc(int wc)" */
    int getwc() {
      return wcount;
    } /* end function: "getwc()" */

    void setinvoc(char *inv) {
      if (inv) strncpy(invocation, inv, MAXLINE);
    } /* end function: "setinvoc(char *inv)" */
    char *getinvoc() {
      return invocation;
    } /* end function: "*getinvoc()" */

    void setnext(envrec *nxt) {
      next = nxt;
    } /* end function: "setnext(envrec *nxt)" */
    envrec *getnext() {
      return next;
    } /* end function: "*getnext()" */
};





                                        /************************************/
                                        /* This sub acquires the DOS master */
                                        /* environment block and inits a    */
                                        /* structure with a pointer to the  */
                                        /* block, and a block size indicator*/
                                        /* Must use a large or compact mem  */
                                        /* model so that pointers can reach */
                                        /* any segment.                     */
                                        /* Comments indicate correspondences*/
                                        /* between assembly source code and */
                                        /* a C imitation of it.  See the    */
                                        /* ASM source code fragment included*/
                                        /* with this package.               */
                                        /************************************/
envrec *get_master_block() {
  union REGS        to_dos;
  union REGS        from_dos;
  struct SREGS      dos_segs;
  void              *tmp_fptr           = NULL;
  unsigned          ax_reg              = 0;
  unsigned          bx_reg              = 0;
  unsigned          di_reg              = 0;
  unsigned          es_reg              = 0;
  unsigned          esize               = 128;
  int               found               = 0;
  int               all_found           = 0;
  envrec            *local_master       = NULL;
  envrec            *tmp_ptr            = NULL;
  envrec            *tmp2_ptr           = NULL;
  char              *first_string       = NULL;
  
                                        /* call intdosx to get the register */
                                        /* values needed to find the DOS    */
                                        /* MASTER environment               */
  to_dos.h.ah = 0x52;
  dos_segs.es = 0;
  intdosx(&to_dos, &from_dos, &dos_segs);
  es_reg = dos_segs.es;
                                        /* Zortech doesn't document it, but */
                                        /* souce code for intdosx shows that*/
                                        /* the .es field in the SREGS var   */
                                        /* is set to the value of the ES    */
                                        /* register before return to calling*/
                                        /* routine.  If not for this fact,  */
                                        /* a substitute for intdosx would   */
                                        /* have to be coded, perhaps as an  */
                                        /* ASM module.  The value in ES is  */
                                        /* pivotal in finding the MCBs.     */
  bx_reg = from_dos.x.bx - 2;
  tmp_fptr = MK_FP(es_reg, bx_reg);
  di_reg = (unsigned)*tmp_fptr;         /* mov  di,es:[bx - 2]              */
  es_reg = di_reg;                      /* mov  es,di                       */
  di_reg += 1;                          /* inc  di                          */

  tmp_fptr = MK_FP(es_reg, 3);
  di_reg += (unsigned)*tmp_fptr;        /* add  di,es:[3]                   */
  es_reg = di_reg;                      /* mov  es,di                       */
  di_reg += 1;                          /* inc  di                          */
  while (!found)                        /* GetEnv:                          */
  {
    tmp_fptr = MK_FP(es_reg, 3);
    di_reg += (unsigned)*tmp_fptr;      /* add  di,es:[3]                   */
    es_reg = di_reg;                    /* mov  es,di                       */
    di_reg += 1;                        /* inc  di                          */

    tmp_fptr = MK_FP(es_reg, 1);
    found = ((unsigned)*tmp_fptr != 0); /* cmp	word ptr es:[1],0 					*/
                                        /* jz 	GetEnv											*/
  } /* end "while (!found)" */
  tmp_fptr = MK_FP(es_reg, 3);
  ax_reg = (unsigned)*tmp_fptr;         /* mov  ax,es:[3]                   */
  ax_reg <<= 4;                         /* shl  ax,4                        */

  esize = ax_reg;
  local_master = new envrec(MK_FP(di_reg, 0), ax_reg);
  first_string = (char*)local_master->getloc();
                                        /* Now first_string points to first */
                                        /* environment string of master env */
                                        /* Must search for subsequent copies*/
                                        /* of environment.  The first string*/
                                        /* is used as a double check that   */
                                        /* we are indeed looking at an env  */
                                        /* block.                           */
  all_found = false;
                                        /* All_found will flag that last env*/
                                        /* copy has been located.           */
  found = false;
  while (!all_found) {                  /* GetEnv:													*/
    tmp_fptr = MK_FP(es_reg, 0);
    if (*(UCHAR*)tmp_fptr > 0x5A) all_found = true;
    else {
      tmp_fptr = MK_FP(es_reg, 3);
      di_reg += (unsigned)*tmp_fptr;    /* add  di,es:[3]                   */
      es_reg = di_reg;                  /* mov  es,di                       */
      di_reg += 1;                      /* inc  di                          */
      tmp_fptr = MK_FP(di_reg, 0);
      found = (strcmp((char*)tmp_fptr, first_string) == 0);
                                        /* cmp  word ptr es:[1],0           */
                                        /* jz   GetEnv                      */
      if (found) {
        tmp_fptr = MK_FP(es_reg, 3);
        ax_reg = (unsigned)*tmp_fptr;   /* mov  ax,es:[3]                   */
        ax_reg <<= 4;                   /* shl  ax,4                        */
                                        /* Don't copy environments smaller  */
                                        /*  than master, because these      */
                                        /*  presumably cannot grow without  */
                                        /*  clobbering something vital.     */
        if (ax_reg >= esize) {
          tmp2_ptr = new envrec(MK_FP(di_reg, 0), ax_reg);
          if (tmp_ptr == NULL) local_master->setnext(tmp2_ptr);
          else tmp_ptr->setnext(tmp2_ptr);
          tmp_ptr = tmp2_ptr;
        } /* end "if (ax_reg >= esize)" */
      } /* end "if (found)" */
    } /* end else clause of "if (*(UCHAR*)tmp_fptr > 0x5A)" */
  } /* end "while (!all_found)" */
  return local_master;
} /* end function: "*get_master_block()" */



                                        /************************************/
                                        /* This proc copies an environment  */
                                        /* block into local storage for ease*/
                                        /* of manipulation.                 */
                                        /* It returns a pointer to the 2nd  */
                                        /* byte following the last string in*/
                                        /* the block.                       */
                                        /* For a discussion of the layout of*/
                                        /* the environment block, see p224  */
                                        /* Advanced MSDOS Programming       */
                                        /*          by Ray Duncan           */
                                        /* (He calls MCBs Arena Headers)    */
                                        /************************************/
void *get_strings(envrec *eptr) {
  char              *str                = NULL;
  void              *tmp_fptr           = NULL;
  unsigned int      segptr;
  unsigned int      offset;
  unsigned int      slen;
  
  str = (char*)eptr->getloc();
  while (strlen(str) > 0) {
    eptr->addvar(str);
    str += (long)(strlen(str) + 1);
  } /* end "while (strlen(str) > 0)" */
  str += (long)(strlen(str) + 1);
  tmp_fptr = (void*)str;
  return tmp_fptr;
} /* end function: "*get_strings(envrec *eptr)" */



                                        /************************************/
                                        /* This is needed by qsort() funct  */
                                        /************************************/
int comp(UCHAR **a, UCHAR **b) {
  return strcmp((char*)*a, (char*)*b);
} /* end function: "comp(UCHAR **a, UCHAR **b)" */



                                        /************************************/
                                        /* Sorting is a frill               */
                                        /************************************/
void sort_block(envrec *eptr) {
  UCHAR             *ptrblk[256];
  slist             *tslist             = NULL;
  unsigned int      nel                 = 0;
  
  tslist = eptr->getlist();
  while (tslist) {
    ptrblk[nel++] = (UCHAR*)tslist->getvar();
    tslist = tslist->getnext();
    ptrblk[nel] = NULL;
  } /* end "while (tslist)" */
  qsort(ptrblk, nel, sizeof (UCHAR*), comp);
  tslist = eptr->getlist();
  nel = 0;
  while (tslist) {
    tslist->setvptr((char*)ptrblk[nel]);
    nel += 1;
    tslist = tslist->getnext();
  } /* end "while (tslist)" */
} /* end function: "sort_block(envrec *eptr)" */



                                        /************************************/
                                        /* This proc writes the blocks back */
                                        /* to memory.                       */
                                        /************************************/
int put_strings(envrec *eptr) {
  int               local_status        = SUCCEEDED;
  void              *gen_ptr            = NULL;
  void              *end_of_block       = NULL;
  char              *gen_str            = NULL;
  slist             *gen_slist          = NULL;
  
  gen_ptr = eptr->getloc();             /* pointer back to start of eptr    */
                                        /* environment                      */
  end_of_block = gen_ptr + (long)eptr->getsize();
                                        /* keep an overrun check value      */
  gen_slist = eptr->getlist();          /* Now replace the environment      */
                                        /* strings                          */
  while ((gen_slist) && (local_status == SUCCEEDED)) {
    gen_str = gen_slist->getvar();
    if ((gen_ptr + (long)strlen(gen_str) + 1 + sizeof (int)) > end_of_block) {
      fprintf(stderr, "put_strings( %d ) - Out of environment space in envset!\n", __LINE__);
      local_status = FAILED;
    } /* end "if ((gen_ptr + (long)strlen(gen_str) + 1 + sizeof (int)) > end_of_block)" */
    else {
      if (strlen(gen_str) > 0) {
        strcpy((char*)gen_ptr, gen_str);/* write a variable to env block    */
        gen_ptr += (long)strlen(gen_str) + 1;
                                        /* bump env block pointer           */
      } /* end "if (strlen(gen_str) > 0)" */
      gen_slist = gen_slist->getnext();
    } /* end else clause of "if ((gen_ptr + (long)strlen(gen_str) + 1 + sizeof (int)) > end_of_block)" */
  } /* end "while ((gen_slist) && (local_status == SUCCEEDED))" */
                                        /* Now add second null to terminate */
                                        /* the string series.               */
  *(char*)gen_ptr = '\0';
  gen_ptr += (long)1;
  if (local_status == SUCCEEDED) {
    *(int*)gen_ptr = eptr->getwc();     /* Replace the word count           */
    gen_ptr += (long)sizeof (int);      /* bump past word count             */
    if ((gen_ptr + (long)strlen(eptr->getinvoc())) > end_of_block) {
      fprintf(stderr, "put_strings( %d ) - Out of environment space in envset!\n", __LINE__);
      local_status = FAILED;
    } /* end "if ((gen_ptr + (long)strlen(eptr->getinvoc())) > end_of_block)" */
    else strcpy((char*)gen_ptr, eptr->getinvoc());
  } /* end "if (local_status == SUCCEEDED)" */
  return local_status;
} /* end function: "put_strings(envrec *eptr)" */



                                        /************************************/
                                        /* Dress this up with ansi escape   */
                                        /* sequences if you are using the   */
                                        /* ANSI.SYS screen driver           */
                                        /************************************/
void usage() {
  fprintf(stderr, "USAGE:\n\tSEV env_var [env_var . . .]\n\n");
  fprintf(stderr, "Where the env_var params specify environment variable to set.\n");
  fprintf(stderr, "\tThe values to be placed in them are derived from standard\n");
  fprintf(stderr, "input.  The purpose of SEV is to allow keyboard input, or the\n");
  fprintf(stderr, "output of other programs, to be captured as environment variables\n");
  fprintf(stderr, "for subsequent use during batch file processing.\n");
  fprintf(stderr, "\n");
} /* end function: "usage()" */



                                                                               
int main(int argc, char **argv) {
  int               status              = SUCCEEDED;
  envrec            *master             = NULL;
  envrec            *e_rec_ptr          = NULL;
  void              *generic_ptr        = NULL;
  char              var_in[MAXLINE];
  char              newstr[255];
  int               count               = 1;
  
  if ((argc == 1) || (argv[1][0] == '?')) {
    usage();
    status = FAILED;
  } /* end "if ((argc == 1) || (argv[1][0] == '?'))" */
  else {
    master = get_master_block();
                                        /* At this point we have acquired   */
                                        /*  the DOS Master environment block*/
                                        /*  and all subsequent environment  */
                                        /*  blocks.                         */
                                        /* Now copy the strings into local  */
                                        /*  storage for manipulation.       */
    e_rec_ptr = master;
                                        /* Loop through list of blocks found*/
    while (e_rec_ptr) {
      generic_ptr = get_strings(e_rec_ptr);
                                        /* Got env variables                */
      e_rec_ptr->setwc((int)*generic_ptr);
                                        /* Got following word count         */
      generic_ptr += (long)sizeof (int);
      e_rec_ptr->setinvoc((char*)generic_ptr);
                                        /* Got the following invocation line*/
      e_rec_ptr->compute_used();
                                        /* Determine how much of env in use */
      e_rec_ptr = e_rec_ptr->getnext();
    } /* end "while (e_rec_ptr)" */
                                        /* At this point a full copy of   */
                                        /* the environment is in memory   */
                                        /* Now add the new strings to it. */
    while ((count < argc) && (status == SUCCEEDED)) {
      if (feof(stdin)) status = FAILED;
      else {
        strcpy(newstr, strupr(argv[count]));
        strcat(newstr, "=");
        var_in[0] = '\0';
        fgets(var_in, MAXLINE, stdin);
        if (var_in[0] != '\0') var_in[strlen(var_in) - 1] = '\0';
        strcat(newstr, var_in);
        e_rec_ptr = master;
        while (e_rec_ptr) {
          e_rec_ptr->addvar(newstr);
          e_rec_ptr = e_rec_ptr->getnext();
        } /* end "while (e_rec_ptr)" */
        count += 1;
      } /* end else clause of "if (feof(stdin))" */
    } /* end "while ((count < argc) && (status == SUCCEEDED))" */
                                        /* If OK here, then write the stuff */
                                        /* back into DOS.                   */
    if (status == SUCCEEDED) {
      e_rec_ptr = master;
      while ((e_rec_ptr) && (status == SUCCEEDED)) {
        sort_block(e_rec_ptr);          /* Why?   Why not!                  */
        status = put_strings(e_rec_ptr);
        e_rec_ptr = e_rec_ptr->getnext();
      } /* end "while ((e_rec_ptr) && (status == SUCCEEDED))" */
    } /* end "if (status == SUCCEEDED)" */
  } /* end else clause of "if ((argc == 1) || (argv[1][0] == '?'))" */
  exit((status == SUCCEEDED) ? 0 : 1);
} /* end function: "main(int argc, char **argv)" */

